# CommandLine.py - Object oriented command line interface.
# vim:sw=4:ts=4:et:nowrap

import os, sys, string
from distutils.fancy_getopt import wrap_text

TERMINAL_WIDTH=78
INDENT=2

class ParseError(Exception):
    """Exception thrown by option and argument parsing errors."""
    pass

class ArgumentError(Exception):
    pass

class OptionError(Exception):
    pass

class Argument:

    TYPES = ['str', 'int', 'list']
    OCCURS = ['once', 'zero', 'oneormore', 'zeroormore']

    def __init__(self, name, type='str', key=None, 
                 descr=None, list=None, occurs='once'):
        self.name = name       # Argument name
        self.type = type       # Argument type (boolean, str, int, list)
        self.key = key         # Unique argument key
        self.descr = descr     # Short description
        self.list = list       # List for validation of 'list' types.
        self.occurs = occurs   # Occurances of argument type.
        self.value = None

    def __repr__(self):
        return '%s' % (self.name)

    def _validate_value(self, value):
        if self.type == 'str':
            return 1
        elif self.type == 'int':
            try:
                int(value)
            except ValueError:
                return 0
            return 1
        elif self.type == 'list':
            if value in self.list:
                return 1
        else:
            return 0

    def usage(self):
        if self.occurs == 'once':
            return '%s' % self.name
        elif self.occurs == 'zero':
            return '[%s]' % self.name
        elif self.occurs == 'oneormore':
            return '%s...' % self.name
        elif self.occurs == 'zeroormore':
            return '[%s]...' % self.name
        else:
            return None

    def validate(self):
        if self.key == None:
            raise ArgumentError,'No key provided for arg %s' % self
        if not self.type in self.TYPES:
            raise ArgumentError,'Bad type %s in arg %s' % (self.type, self)
        if not self.occurs in self.OCCURS:
            raise ArgumentError,'Bad occurance %s arg %s' % (self.occurs, self)
        if self.type == 'list' and self.list == None:
            raise ArgumentError,'No validation list provided for %s' % self

    def set_value(self, value):
        if self._validate_value(value):
            self.value = value
        else:
            raise ArgumentError,'%s is an invalid value for %s' % (value, self)

class Option:

    TYPES = ['boolean', 'str', 'int', 'list']

    def __init__(self, short, long, type='boolean', key=None, 
                 descr=None, argname=None, list=None):
        self.short = short     # Short option flag (--o)
        self.long = long       # Long option flag ( --option)
        self.type = type       # Option type (boolean, str, int, list)
        self.key = key         # Unique option key
        self.descr = descr     # Short description
        self.argname = argname # Name of argument for option
        self.list = list       # List for validation of 'list' types.
        self.value = None

    def __repr__(self):
        return '%s/%s' % (self.short, self.long)

    def _validate_value(self, value):
        if self.type == 'boolean' or self.type == 'str':
            return 1
        elif self.type == 'int':
            try:
                int(value)
            except ValueError:
                return 0
            return 1
        elif self.type == 'list':
            if value in self.list:
                return 1
        else:
            return 0

    def usage(self):
        if self.type == 'boolean':
            return '[%s]' % self.short
        else:
            return '[%s %s]' % (self.short, self.argname)

    def set_value(self, value):
        if self._validate_value(value):
            self.value = value
        else:
            raise OptionError,'%s is an invalid value for %s' % (value, self)

    def validate(self):
        # Short options must take the form -x
        if not len(self.short) == 2 or not self.short.startswith('-') or \
        self.short[1] == '-':
            raise OptionError, 'Illegal option %s' % self.short
        # Long options must take the form --xx...
        if not len(self.long) > 4 or not self.long.startswith('--') or \
        self.long[2] == '-':
            raise OptionError, 'Illegal option %s' % self.short
        if self.key == None:
            raise OptionError,'No key provided for option %s' % self
        if not self.type in self.TYPES:
            raise OptionError,'Bad type %s in option %s' % (self.type, self)
        if self.type == 'list' and self.list == None:
            raise OptionError,'No validation list provided for %s' % self
        elif self.type != 'boolean' and self.argname == None:
            raise OptionError,'No argument name provided for %s' % self

class Command:
    """
    Class attrs:
      GLOBALOPTS : [Options.Option]
      List of options included in every Command class instance, and
      therefore global to every command.

    Public instance attrs:
      name : string
      The name of the program that the command instance is invoked
      from. (sys.argv[0])
      help : string
      Optional description of the program and its purpose that will
      be included in the usage. (__doc__)
      options : [Options.Option]
      list of options accepted by the command instance.
      arguments : [Arguments.Argument]
      list of positional arguments accepted by the command instance
    """

    GLOBALOPTS = [ Option('-h','--help',key='help',
                         descr='output this message and exit'),]

    def __init__(self, name, help=None, options=None, arguments=None):
        self.name = os.path.split(name)[1]
        self.help = help
        self.options = options or []
        self.arguments = arguments or []
        self._usage = None
        self.options = self.GLOBALOPTS + self.options
        self._validate_commandline()
        self._generate_usage()

    def _validate_commandline(self):
        [a.validate() for a in self.options + self.arguments]
        # TODO Check for positional argument issues

    def _generate_usage(self):
        usage = [self.name]
        usage.extend([a.usage() for a in self.options + self.arguments])
        self._usage = string.join(usage)

    def _parse_error(self, err):
        print >> sys.stderr, '%s Fatal: %s' % (self.name, err)
        self._print_usage()
        sys.exit(127)

    def _print_usage(self):
        width = TERMINAL_WIDTH - INDENT
        if self.help:
            print >> sys.stdout
            for line in wrap_text(self.help, width):
                print >> sys.stdout, "%*s%s" % (INDENT,'', line)
        print >> sys.stdout
        print >> sys.stdout, 'Usage:'
        for line in wrap_text(self._usage, width):
            print >> sys.stdout, '%*s%s' % (INDENT,'', line)
        print >> sys.stdout
        # Print out a nicely formatted Options table, if there are options.
        if self.options:
            maxlen = 0
            optdict = {}
            for opt in self.options:
                optstr = ''
                if opt.type == 'boolean':
                    optstr = "%*s%s" % (INDENT, '', opt)
                else:
                    optstr = "%*s%s '%s'" % (INDENT, '' , opt, opt.argname)
                if len(optstr) > maxlen: maxlen = len(optstr)
                if opt.type == 'list':
                    opt.descr = '%s (%s)' % (opt.descr, 
                                             string.join(opt.list,', '))
                optdict[opt.key] = [optstr, opt.descr]
            print >> sys.stdout, 'Options:'
            for optkey in optdict.keys():
                optstr, optdescr  = optdict[optkey]
                if len(optstr) <= maxlen:
                    optstr = '%s%s' %  (optstr, ' ' * ( maxlen - len(optstr)))
                print >> sys.stdout, '%s\t' % optstr,
                line  = wrap_text(optdescr, width - (maxlen + 7))
                print >> sys.stdout, line[0]
                for l in line[1:]:
                    print >> sys.stdout, '%*s\t%s' % (INDENT + maxlen,'',l)
        print >> sys.stdout
        # Print out a nicely formatted Arguments table, if there are arguments.
        if self.arguments:
            maxlen = 0
            argdict = {}
            for arg in self.arguments:
                argstr = ''
                argstr = "%*s%s" % (INDENT, '', arg.name)
                if len(argstr) > maxlen: maxlen = len(argstr)
                if arg.type == 'list':
                    arg.descr = '%s (%s)' % (arg.descr, 
                                             string.join(arg.list,', '))
                argdict[arg.key] = [argstr, arg.descr]
            print >> sys.stdout, 'Arguments:'
            for argkey in argdict.keys():
                argstr, argdescr  = argdict[argkey]
                if len(argstr) <= maxlen:
                    argstr = '%s%s' %  (argstr, ' ' * ( maxlen - len(argstr)))
                print >> sys.stdout, '%s\t' % argstr,
                line  = wrap_text(argdescr, width - (maxlen + 7))
                print >> sys.stdout, line[0]
                for l in line[1:]:
                    print >> sys.stdout, '%*s\t%s' % (INDENT + maxlen,'',l)

    def _parse(self):
        rargs = sys.argv[1:]
        rargs.reverse()
        pargs = []
        while rargs:
            arg = rargs.pop()
            if arg == '--':
                # Stop processing arguments.
                pargs.extend(rargs)
                break
            elif arg == '-h' or arg == '--help':
                self._print_usage()
                sys.exit(0)
            elif arg[0:2] == '--':
                # Process long option
                if '=' in arg:
                    # If option value is set explicitly (--foo=bar) then
                    # insert the value into rargs and make a not of it.
                    (opt, next_arg) = arg.split('=', 1)
                    rargs.insert(0, next_arg)
                    explicit_value = 1
                else:
                    opt = arg
                    explicit_value = 0
                if self.has_opt(opt):
                    # Check to see that the Command instance has a option of
                    # the type 'opt registered.
                    option = self.get_opt(opt)
                else:
                    raise ParseError,'no such option, %s' % opt
                if option.type != 'boolean':
                    if len(rargs) < 1:
                        raise ParseError,'option %s requires a value' % opt
                    else:
                        option.set_value(rargs.pop())
                else:
                    if explicit_value:
                        raise ParseError,"option %s doesn't take a value" % opt
                    option.set_value(1)
                # End long option processing
            elif arg[:1] == '-' and len(arg) > 1:
                # Process short option
                i = 1
                for ch in arg[1:]:
                    opt = '-' + ch
                    # Check to see that the option exists.
                    if self.has_opt(opt):
                        option = self.get_opt(opt)
                        i += 1
                    else:
                        raise ParseError, 'no such option, %s' % opt

                    if option.type != 'boolean':
                        if i < len(arg):
                            rargs.insert(0, arg[i:])
                        if len(rargs) < 1:
                            raise ParseError, 'option %s requires a value' % opt
                        else:
                            option.set_value(rargs.pop())
                    else:
                        option.set_value(1)
            else:
                pargs.append(arg)
        if self.arguments:
            pargs.reverse()
            for a in self.arguments:
                try:
                    parg = pargs.pop()
                except IndexError:
                    parg = None
                if a.occurs == 'once':
                    if parg:
                        try:
                            a.set_value(parg)
                        except Exception, err:
                            raise ParseError, err
                    else:
                        raise ParseError,"Missing required argument '%s'" % a
                    continue
                elif a.occurs == 'zero':
                    if parg:
                        try:
                            a.set_value(parg)
                        except Exception, err:
                            raise ParseError, err
                    continue
                elif a.occurs == 'oneormore': 
                    if parg:
                        pargs.insert(len(pargs)+1,parg)
                        pargs.reverse()
                        try:
                            a.set_value(pargs)
                            pargs = []
                        except Exception, err:
                            raise ParseError, err
                    else:
                        raise ParseError,"Missing required argument '%s'" % a
                    break
                elif a.occurs == 'zeroormore':
                    if parg:
                        pargs.insert(len(pargs)+1,parg)
                        pargs.reverse()
                        try:
                            a.set_value(pargs)
                            pargs = []
                        except Exception, err:
                            raise ParseError, err
                    continue
        if pargs:
            raise ParseError, "Nothing to do with %s" % string.join(pargs, ',')

    def has_opt(self, opt):
        for option in self.options:
            if option.short == opt or option.long == opt:
                return 1
        return 0

    def get_opt(self, opt):
        for option in self.options:
            if option.short == opt or option.long == opt:
                return option
        return None

    def has_arg(self, arg):
        for argument in self.arguments:
            if argument.name == arg:
                return 1
        return 0

    def get_arg(self, arg):
        for argument in self.arguments:
            if argument.name == arg:
                return arg
        return None

    def has_value(self, key):
        pass

    def get_value(self, key):
        for i in self.options + self.arguments:
            if i.key == key:
                return i.value
            else:
                continue
        return None

    def parse(self):
        try:
            self._parse()
        except ParseError, err:
            self._parse_error(err)

