#
# scast-util.py - Utility functions common to scast installs, upgrades
#

import os
import re
import sys
import string
import logging
import shutil
import tempfile
import subprocess
import select
import fcntl
import commands
import time
import glob
from RelengErrUtil import *
from CommonUtils import CommonUtils
from ConfigParser import ConfigParser

# Variables uses throughout the scast-*.py scripts

TRUE = 1
FALSE = 0
DEBUG = 0
ENVCONFIG = '.env-conf'
YES_NO_OPTIONS = ['NO','YES']
JVM_DEFAULT_OPTIONS = '-Dfile.encoding=UTF-8 -verbose:gc -XX:+PrintGCTimeStamps -XX:+PrintGCDetails'
INSTALLER_CONF='.installer-conf'
ENV_CONF='.env-conf'

DATABASETOKENS = {
           'postgres' : {
                'dbusername': 'DATABASE_USERNAME',
                'port': 'DATABASE_PORT',
                'dbpassword' : 'DATABASE_PASSWORD',
                'dbname': 'DATABASE_NAME',
                },
            'oracle' : {
                'port' : 'REPORTING_DATABASE_PORT',
                },
             }


class InstallerUtil(CommonUtils):
 
        def executeOsSystemCommand(self, command):
            """ 
            Executes the command, by running it as a subprocess
            Returns the content of the std out stream if no error occurs else return
            the error stream content and raise a runtime exception
            """
            # create a child process, we want the stderr as a sepearte stream
            child = subprocess.Popen(command, shell=True, stdout=subprocess.PIPE, stdin=subprocess.PIPE, stderr=subprocess.PIPE)
            childOut, childErr = child.communicate()
            sys.stdout.write (childOut)
            err = child.returncode
            if err != 0:
               raise RelengErr(make_str(802,command,err,childErr))
            return err,childOut,childErr

        def executeCommandInteractiveRaw(self, command):
            """ 
            Executes the command, by running it as a subprocess and displays the 
            stdout/stderr stream interactively.
            Returns the content of the std out stream if no error occurs else return
            the error stream content and raise a runtime exception
            Note: This function is to be used only for those commands that produces
            lot of outputs while running.
            """
            # create a child process, we want the stderr as a sepearte stream
            child = subprocess.Popen(command, shell=True, stdout=subprocess.PIPE, stdin=None, stderr=subprocess.PIPE)
            childOut = ''
            childErr = ''
            while True:
                output_line = self.readProcData(child.stdout)
                child.stdout.flush()
                sys.stdout.write(output_line)
                sys.stdout.flush()
                if child.poll() != None:
                   break

            while True:
                output_line = child.stderr.readline()
                if output_line == '' and child.poll() != None:
                    break
                sys.stderr.write(output_line)
                sys.stderr.flush()
                childErr += output_line

            err = child.returncode
            return err,childOut,childErr

        def executeOsSystemCommandInteractive(self, command):
            """ 
            Executes the command, by running it as a subprocess and displays the 
            stdout/stderr stream interactively.
            Returns the content of the std out stream if no error occurs else return
            the error stream content and raise a runtime exception
            Note: This function is to be used only for those commands that produces
            lot of outputs while running.
            """
            # create a child process, we want the stderr as a sepearte stream
            child = subprocess.Popen(command, shell=True, stdout=subprocess.PIPE, stdin=None, stderr=subprocess.PIPE)
            childOut = ''
            childErr = ''
            while True:
                output_line = self.readProcData(child.stdout)
                child.stdout.flush()
                sys.stdout.write(output_line)
                sys.stdout.flush()
                if child.poll() != None:
                   break

            while True:
                output_line = child.stderr.readline()
                if output_line == '' and child.poll() != None:
                    break
                sys.stderr.write(output_line)
                sys.stderr.flush()
                childErr += output_line

            err = child.returncode
            if err != 0:
               raise RelengErr(make_str(802,command,err,childErr))
            return err,childOut,childErr

        def readProcData(self, fObj):
    
            (r, w, x) = select.select([fObj], [], [], 0.2)
            value = ""
            while r:
                char = fObj.read(1)
                if (len(char) == 0):
                   break
                value += char;
                (r, w, x) = select.select([fObj], [], [], 0.2)
    
            return value;
    
        def BootstrapData(self, siteDir, flags = {}, yes = False):
            logFile = os.path.normpath(os.path.join(siteDir, 'log', 'runtime', 'bootstrap.log'))
            exec_string = os.path.join(siteDir, 'runtime', 'scripts', 'bootstrap-data.py')
            if yes:
                exec_string += " -y"
            self.out('Bootstrapping the data...')
            return_code = os.system(exec_string)
            return return_code

        def createRuntime(self, siteDir, siteOptionsFile, skip_verify = False, yes = False):
            logFile = os.path.normpath(os.path.join(siteDir, 'log', 'runtime', 'runtime.log'))
            """Create runtime configuration"""
            exec_string = os.path.join(siteDir, 'dist', 'scripts',
                                        'create-runtime.py')
            exec_string = exec_string + " -f " + siteOptionsFile
            if skip_verify:
                exec_string += " -s"
            if yes:
                exec_string += " -y"

            self.out('\nConfiguring the system with your options ... please be patient ...')

            return_code = os.system(exec_string)
            return return_code

        def readFromUser(self, prompt, default, valid_values = 'y/N'):
            """Get input from the user with a default value and pretty printing"""
            ret_val = ''
            i = 0

            while ret_val == '':
                if i == 0:
                    ret_val =  raw_input(prompt + '  [' + default + ']  ')
                else:
                    ret_val =  raw_input('  [' + default + ']  ')

                if default != '' and ret_val == '' :
                    ret_val = default
                elif default == '' and ret_val == '':
                    raise RelengErr(make_str(821, valid_values))

                if valid_values and not ret_val in valid_values:
                    RelengErr (make_str(821, valid_values))
                    ret_val = ''
                    i += 1
            return ret_val

        def islocalhost(self, hostname):
            islocalhost = 1
            if hostname != 'localhost':
                if (socket.gethostbyname(hostname) !=
                    socket.gethostbyname(socket.gethostname())):
                    islocalhost = 0
            return islocalhost

